/*==============================================================================*/
/*  Copyright (C) 2020, Heartland Data inc. All Rights Reserved.                */
/*                                                                              */
/*  Title  :   Data Writing Driver                                              */
/*  EventID:   Via the driver                                                   */
/*  AppVer :   DT+ Ver1.00～                                                    */
/*  FileID :   38c1                                                             */
/*  Version:   1.0                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Desc:   Header for File IO                                                  */
/*==============================================================================*/
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>

/*==============================================================================*/
/*  Macro:  DT_UINT                                                             */
/*  Desc:   Please change Test Point argument type for DT+ Project setting.     */
/*==============================================================================*/
#define	DT_UINT unsigned int

/*==============================================================================*/
/*  Desc:   Test Point Parametaer                                               */
/*==============================================================================*/
#define DT_VARIABLE_BIT		0x2

/*==============================================================================*/
/*  Macro:  DT_ADD_EVENT_INFO                                                   */
/*  Desc:   Please set 1 when you add Event Value to Test Point.                */
/*==============================================================================*/
#define DT_ADD_EVENT_INFO	1

static FILE *dat_file = NULL;
static int BusoutMode = 0;
unsigned int pid = 0;
static unsigned int start_time;

/*==============================================================================*/
/*  Desc:   mutex                                                               */
/*==============================================================================*/

/* TP write mutex */
pthread_mutex_t mutex;
#define MUTEX_INIT	pthread_mutex_init(&mutex, NULL)
#define MUTEX_WAIT	pthread_mutex_lock(&mutex)
#define MUTEX_POST	pthread_mutex_unlock(&mutex)

/* file write mutex */
pthread_mutex_t mutex_write;

/*==============================================================================*/
/*  Macro:  BusOutModeSet                                                       */
/*  Desc:   0 => open file 1 => writing 2 => close file                         */
/*==============================================================================*/
#define BusOutModeSet(mode) BusoutMode = mode

/*==============================================================================*/
/*  Desc:   Please set buffer size                                              */
/*==============================================================================*/
#define DT_MAX_TESTREPORT_NUMBER	50000	/* Customize */
#define DT_BUFF_SIZE	(DT_MAX_TESTREPORT_NUMBER * 16)
static unsigned char buff[DT_BUFF_SIZE];
static unsigned int write_pos = 0, read_pos = 0;
static unsigned int write_len = 0, read_len = 0;
pthread_t thread;


/* 1 -> use ring buffer */
#define DT_USE_RING_BUFF		1
/* buffer overflow if end trace */
#define DT_BUFF_OVERFLOW_TRACE_END	0

void *_TP_WriteFileThread(void *arg)
{
	unsigned int tmp;
	/* Customize File Path */
	if (dat_file == NULL) {
		dat_file = fopen("ConvertData.DAT", "wb+");
		if (dat_file == NULL) {
			return;
		}
	}
	while (1) {
		pthread_mutex_lock(&mutex_write);
		if (read_pos + write_len < DT_BUFF_SIZE) {
			tmp = read_pos + write_len;
			fwrite(buff+read_pos, sizeof(char), write_len, dat_file);
			read_pos += tmp;
			read_len += write_len;
		} else {
			tmp = DT_BUFF_SIZE - read_pos;
			read_len += tmp;
			fwrite(buff+read_pos, sizeof(char), tmp, dat_file);
#if DT_USE_RING_BUFF
			read_pos = 0;
			fwrite(buff+read_pos, sizeof(char), (write_len - tmp), dat_file);
#else
			pthread_mutex_unlock(&mutex_write);
			break;
#endif
		}
		pthread_mutex_unlock(&mutex_write);
		write_len = 0;
		sleep(1);
	}
	fclose(dat_file);
	pthread_exit(NULL);
}

static inline void _TP_BusOutInit(void)
{
	struct timespec now;

	MUTEX_INIT;
	pthread_mutex_init(&mutex_write, NULL);
	if (pthread_create(&thread, NULL, &_TP_WriteFileThread, NULL) != 0) {
		return;
	}
	/* start time data */
	if (clock_gettime(CLOCK_MONOTONIC, &now) == -1) {
		printf("DT : timer error!\n");
	}
	start_time = (unsigned int)((now.tv_sec*1000*1000) + (now.tv_nsec/1000));
	BusOutModeSet(1);
}

static inline void _TP_BusOutDrv(DT_UINT addr, DT_UINT dat)
{
	struct timespec now;
	unsigned int time;

#if (DT_USE_RING_BUFF== 0)
	if (write_pos >= DT_BUFF_SIZE) {
		return;
	}
#endif

#if DT_ADD_EVENT_INFO
	pid = getpid();
#endif
	/* time data */
	if (clock_gettime(CLOCK_MONOTONIC, &now) == -1) {
		printf("DT : timer error!\n");
	}
	time = (unsigned int)((now.tv_sec*1000*1000) + (now.tv_nsec/1000));
	time -= start_time;

	buff[write_pos++] = (unsigned char)((dat) & 0xFF);
	buff[write_pos++] = (unsigned char)((dat >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((addr) & 0xFF);
	buff[write_pos++] = (unsigned char)((addr >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((addr >> 16) & 0xFF);
	/* time data */
	buff[write_pos++] = (unsigned char)((time) & 0xFF);
	buff[write_pos++] = (unsigned char)((time >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((time >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((time >> 24) & 0xFF);
#if DT_ADD_EVENT_INFO
	/* pid */
	buff[write_pos++] = (unsigned char)((pid) & 0xFF);
	buff[write_pos++] = (unsigned char)((pid >> 8) & 0xFF);
	buff[write_pos++] = (unsigned char)((pid >> 16) & 0xFF);
	buff[write_pos++] = (unsigned char)((pid >> 24) & 0xFF);
#else
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;
#endif
	/* Reserve */
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;
	buff[write_pos++] = 0x00;

	write_len += 16;
#if DT_USE_RING_BUFF
	if (write_pos >= DT_BUFF_SIZE) {
		write_pos = 0;
	}
	/* buffer overflow */
	if (write_len > DT_BUFF_SIZE) {
		/* delete old data */
		read_pos = write_pos;
#if DT_BUFF_OVERFLOW_TRACE_END
		/* end trace */
		BusOutModeSet(2);
#else
		write_len -= 16;
#endif
	}
#endif
}

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
void _TP_BusOut(DT_UINT addr, DT_UINT dat)
{
	if (BusoutMode == 0) {
		_TP_BusOutInit();
	}

	MUTEX_WAIT;
	_TP_BusOutDrv(addr, dat);
	MUTEX_POST;
}

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput(DT_UINT addr, DT_UINT dat, void *value, DT_UINT size)
{
	char *p = value;
	int i, j, packetNum, count;

	if (BusoutMode == 0) {
		_TP_BusOutInit();
	}

	if (size > 256) size = 256;
	packetNum = (size + 7) / 8;

	MUTEX_WAIT;
	_TP_BusOutDrv((addr | DT_VARIABLE_BIT), dat);	/* Memory Dump Start */

	count = 0;
	for (i = 0; i < packetNum; i++) {
#if (DT_USE_RING_BUFF== 0)
		if (write_pos >= DT_BUFF_SIZE) {
			MUTEX_POST;
			return;
		}
#endif
		buff[write_pos++] = size;
		buff[write_pos++] = packetNum;
		buff[write_pos++] = 0x03;
		buff[write_pos++] = i;
		for (j = 0; j < 8; j++, count++, p++) {
			buff[write_pos++] = (count < size)?*p:0x00;
		}
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		buff[write_pos++] = 0x00;
		write_len += 16;
#if DT_USE_RING_BUFF
		if (write_pos >= DT_BUFF_SIZE) {
			write_pos = 0;
		}
		/* buffer overflow */
		if (write_len > DT_BUFF_SIZE) {
			/* delete old data */
			read_pos = write_pos;
#if DT_BUFF_OVERFLOW_TRACE_END
			/* end trace */
			BusOutModeSet(2);
#else
		write_len -= 16;
#endif
		}
#endif
	}
	MUTEX_POST;
}
